@@ -7,7 +7,7 @@ from functools import partial
 from collections import OrderedDict, defaultdict
 from itertools import groupby
 
-from sql import Literal, Union, Column, Null
+from sql import Literal, Union, Column, Null, For
 from sql.aggregate import Sum
 from sql.conditionals import Coalesce
 from sql.operators import Concat
@@ -628,9 +628,6 @@ class Move(Workflow, ModelSQL, ModelView):
     @Workflow.transition('assigned')
     def assign(cls, moves):
         cls.check_origin(moves)
-        for move in moves:
-            move.set_effective_date()
-        cls.save(moves)
 
     @classmethod
     @ModelView.button
@@ -761,7 +758,7 @@ class Move(Workflow, ModelSQL, ModelView):
                 to_pick.append((location, available_qty))
                 needed_qty -= available_qty
         # Force assignation for consumables:
-        if self.product.consumable:
+        if self.product.consumable and self.from_location.type != 'view':
             to_pick.append((self.from_location, needed_qty))
             return to_pick
         return to_pick
@@ -778,8 +775,13 @@ class Move(Workflow, ModelSQL, ModelView):
         Uom = pool.get('product.uom')
         Date = pool.get('ir.date')
         Location = pool.get('stock.location')
+        Period = pool.get('stock.period')
+        transaction = Transaction()
+        database = transaction.database
+        connection = transaction.connection
 
-        Transaction().database.lock(Transaction().connection, cls._table)
+        if not moves:
+            return True
 
         if with_childs:
             locations = Location.search([
@@ -788,12 +790,42 @@ class Move(Workflow, ModelSQL, ModelView):
                     ])
         else:
             locations = list(set((m.from_location for m in moves)))
+        location_ids = [l.id for l in locations]
+        product_ids = list(set((m.product.id for m in moves)))
+        stock_date_end = Date.today()
+
+        if database.has_select_for():
+            table = cls.__table__()
+            query = table.select(Literal(1),
+                where=(table.to_location.in_(location_ids)
+                    | table.from_location.in_(location_ids))
+                & table.product.in_(product_ids),
+                for_=For('UPDATE', nowait=True))
+
+            PeriodCache = Period.get_cache(grouping)
+            if PeriodCache:
+                periods = Period.search([
+                        ('date', '<', stock_date_end),
+                        ('state', '=', 'closed'),
+                        ], order=[('date', 'DESC')], limit=1)
+                if periods:
+                    period, = periods
+                    query.where &= Coalesce(
+                        table.effective_date,
+                        table.planned_date,
+                        datetime.date.max) > period.date
+
+            with connection.cursor() as cursor:
+                cursor.execute(*query)
+        else:
+            database.lock(connection, cls._table)
+
         with Transaction().set_context(
-                stock_date_end=Date.today(),
+                stock_date_end=stock_date_end,
                 stock_assign=True):
             pbl = Product.products_by_location(
-                location_ids=[l.id for l in locations],
-                product_ids=[m.product.id for m in moves],
+                location_ids=location_ids,
+                product_ids=product_ids,
                 grouping=grouping)
 
         def get_key(move, location):
@@ -814,7 +846,6 @@ class Move(Workflow, ModelSQL, ModelView):
                 if move.state == 'staging':
                     success = False
                 continue
-            to_location = move.to_location
             # Keep location order for pick_product
             location_qties = OrderedDict()
             if with_childs:
@@ -865,9 +896,7 @@ class Move(Workflow, ModelSQL, ModelView):
                         move.product.default_uom, round=False)
 
                 from_key = get_key(move, from_location)
-                to_key = get_key(move, to_location)
                 pbl[from_key] = pbl.get(from_key, 0.0) - qty_default_uom
-                pbl[to_key] = pbl.get(to_key, 0.0) + qty_default_uom
         if to_write:
             cls.write(*to_write)
         if to_assign:
@@ -886,7 +915,8 @@ class Move(Workflow, ModelSQL, ModelView):
             stock_date_start: if set return the delta of the stock between the
                 two dates, (ignored if stock_date_end is missing).
             stock_assign: if set compute also the assigned outgoing moves as
-                done.
+                done at the stock_date_end except for delta which is at the
+                planned date.
             forecast: if set compute the forecast quantity.
             stock_destinations: A list of location ids. If set, restrict the
                 computation to moves from and to those locations.
@@ -947,6 +977,9 @@ class Move(Workflow, ModelSQL, ModelView):
                         & (move.planned_date <= context['stock_date_end'])
                         )
                     | (move.effective_date <= context['stock_date_end'])
+                    | (move.state == (
+                            'assigned' if not context.get('stock_date_start')
+                            else ''))
                     )
                 )
             state_date_clause_in = state_date_clause(False)
@@ -965,6 +998,10 @@ class Move(Workflow, ModelSQL, ModelView):
                             & (move.planned_date <= today)
                             )
                         | (move.effective_date <= today)
+                        | (move.state == (
+                                'assigned'
+                                if not context.get('stock_date_start')
+                                else ''))
                         )
                     )
                 | (move.state.in_(['done', 'assigned', 'draft'])
@@ -980,6 +1017,10 @@ class Move(Workflow, ModelSQL, ModelView):
                             (move.effective_date <= context['stock_date_end'])
                             & (move.effective_date >= today)
                             )
+                        | (move.state == (
+                                'assigned'
+                                if not context.get('stock_date_start')
+                                else ''))
                         )
                     )
                 )
