@@ -31,9 +31,37 @@ __all__ = ['ShipmentIn', 'ShipmentInReturn',
     'SupplierRestockingList', 'CustomerReturnRestockingList',
     'InteralShipmentReport']
 
-STATES = {
-    'readonly': "state in ('cancel', 'done')",
-}
+
+def set_employee(field):
+    def decorator(func):
+        @functools.wraps(func)
+        def wrapper(cls, shipments, *args, **kwargs):
+            pool = Pool()
+            User = pool.get('res.user')
+            user = User(Transaction().user)
+
+            result = func(cls, shipments, *args, **kwargs)
+            employee = user.employee
+            if employee:
+                company = employee.company
+                cls.write(
+                    [s for s in shipments
+                        if not getattr(s, field) and s.company == company], {
+                        field: employee.id,
+                        })
+            return result
+        return wrapper
+    return decorator
+
+
+def employee_field(string):
+    return fields.Many2One(
+        'company.employee', string,
+        domain=[('company', '=', Eval('company', -1))],
+        states={
+            'readonly': Eval('state').in_(['done', 'cancel']),
+            },
+        depends=['company', 'state'])
 
 
 class ShipmentIn(Workflow, ModelSQL, ModelView):
@@ -126,7 +154,9 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
         domain=[('company', '=', Eval('company'))], readonly=True,
         depends=['company'])
     origins = fields.Function(fields.Char('Origins'), 'get_origins')
-    number = fields.Char('Numer', size=None, select=True, readonly=True)
+    number = fields.Char('Number', size=None, select=True, readonly=True)
+    received_by = employee_field("Received By")
+    done_by = employee_field("Done By")
     state = fields.Selection([
         ('draft', 'Draft'),
         ('done', 'Done'),
@@ -307,7 +337,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
     def get_incoming_moves(self, name):
         moves = []
         for move in self.moves:
-            if move.to_location.id == self.warehouse.input_location.id:
+            if move.to_location == self.warehouse_input:
                 moves.append(move.id)
         return moves
 
@@ -322,8 +352,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
     def get_inventory_moves(self, name):
         moves = []
         for move in self.moves:
-            if (move.from_location.id ==
-                    self.warehouse.input_location.id):
+            if move.from_location == self.warehouse_input:
                 moves.append(move.id)
         return moves
 
@@ -397,6 +426,9 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
         default = default.copy()
         default['inventory_moves'] = None
         default['incoming_moves'] = None
+        default.setdefault('number')
+        default.setdefault('received_by')
+        default.setdefault('done_by')
         return super(ShipmentIn, cls).copy(shipments, default=default)
 
     @classmethod
@@ -410,7 +442,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
         move.uom = incoming_move.uom
         move.quantity = incoming_move.quantity
         move.from_location = incoming_move.to_location
-        move.to_location = incoming_move.shipment.warehouse.storage_location
+        move.to_location = incoming_move.shipment.warehouse_storage
         move.state = Move.default_state()
         # Product will be considered in stock only when the inventory
         # move will be made:
@@ -464,6 +496,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('received')
+    @set_employee('received_by')
     def receive(cls, shipments):
         Move = Pool().get('stock.move')
         Move.do([m for s in shipments for m in s.incoming_moves])
@@ -472,6 +505,7 @@ class ShipmentIn(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('done')
+    @set_employee('done_by')
     def done(cls, shipments):
         pool = Pool()
         Move = pool.get('stock.move')
@@ -553,6 +587,8 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
             ],
         depends=['state', 'from_location', 'to_location', 'company'])
     origins = fields.Function(fields.Char('Origins'), 'get_origins')
+    assigned_by = employee_field("Assigned By")
+    done_by = employee_field("Done By")
     state = fields.Selection([
         ('draft', 'Draft'),
         ('cancel', 'Canceled'),
@@ -725,6 +761,16 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
         Move.delete([m for s in shipments for m in s.moves])
         super(ShipmentInReturn, cls).delete(shipments)
 
+    @classmethod
+    def copy(cls, shipments, default=None):
+        if default is None:
+            default = {}
+        default = default.copy()
+        default.setdefault('number')
+        default.setdefault('assigned_by')
+        default.setdefault('done_by')
+        return super(ShipmentInReturn, cls).copy(shipments, default=default)
+
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
@@ -751,6 +797,7 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
 
     @classmethod
     @Workflow.transition('assigned')
+    @set_employee('assigned_by')
     def assign(cls, shipments):
         Move = Pool().get('stock.move')
         Move.assign([m for s in shipments for m in s.moves])
@@ -758,6 +805,7 @@ class ShipmentInReturn(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('done')
+    @set_employee('done_by')
     def done(cls, shipments):
         pool = Pool()
         Move = pool.get('stock.move')
@@ -856,7 +904,8 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
                 ('company', '=', Eval('company')),
                 ],
             states={
-                'readonly': ((Eval('state').in_(['waiting', 'done', 'cancel']))
+                'readonly': ((Eval('state').in_(
+                            ['waiting', 'packed', 'done', 'cancel']))
                     | ~Eval('warehouse') | ~Eval('customer')),
                 },
             depends=['state', 'warehouse', 'customer', 'warehouse_output',
@@ -872,7 +921,9 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
                 ],
             states={
                 'readonly': Eval('state').in_(
-                    ['draft', 'packed', 'done', 'cancel']),
+                    ['draft', 'assigned', 'packed', 'done', 'cancel']),
+                'invisible': (
+                    Eval('warehouse_storage') == Eval('warehouse_output')),
                 },
             depends=['state', 'warehouse', 'warehouse_storage',
                 'warehouse_output', 'company']),
@@ -882,6 +933,9 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         readonly=True)
     origins = fields.Function(fields.Char('Origins'), 'get_origins')
     number = fields.Char('Number', size=None, select=True, readonly=True)
+    assigned_by = employee_field("Assigned By")
+    packed_by = employee_field("Packed By")
+    done_by = employee_field("Done By")
     state = fields.Selection([
         ('draft', 'Draft'),
         ('done', 'Done'),
@@ -902,6 +956,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         cls._transitions |= set((
                 ('draft', 'waiting'),
                 ('waiting', 'assigned'),
+                ('waiting', 'packed'),
                 ('assigned', 'packed'),
                 ('packed', 'done'),
                 ('assigned', 'waiting'),
@@ -932,13 +987,18 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
                             'tryton-go-next')),
                     },
                 'pack': {
-                    'invisible': Eval('state') != 'assigned',
+                    'invisible': If(Eval('warehouse_storage')
+                        == Eval('warehouse_output'),
+                        ~Eval('state').in_(['assigned', 'waiting']),
+                        Eval('state') != 'assigned'),
                     },
                 'done': {
                     'invisible': Eval('state') != 'packed',
                     },
                 'assign_wizard': {
-                    'invisible': Eval('state') != 'waiting',
+                    'invisible': ((Eval('state') != 'waiting')
+                        | (Eval('warehouse_storage')
+                            == Eval('warehouse_output'))),
                     },
                 'assign_try': {},
                 'assign_force': {},
@@ -1034,7 +1094,11 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
     @fields.depends('warehouse')
     def on_change_with_warehouse_storage(self, name=None):
         if self.warehouse:
-            return self.warehouse.storage_location.id
+            if self.warehouse.picking_location:
+                location = self.warehouse.picking_location
+            else:
+                location = self.warehouse.storage_location
+            return location.id
 
     @classmethod
     def default_warehouse_output(cls):
@@ -1050,7 +1114,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
     def get_outgoing_moves(self, name):
         moves = []
         for move in self.moves:
-            if move.from_location.id == self.warehouse.output_location.id:
+            if move.from_location == self.warehouse_output:
                 moves.append(move.id)
         return moves
 
@@ -1065,7 +1129,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
     def get_inventory_moves(self, name):
         moves = []
         for move in self.moves:
-            if move.to_location.id == self.warehouse.output_location.id:
+            if move.to_location == self.warehouse_output:
                 moves.append(move.id)
         return moves
 
@@ -1106,6 +1170,9 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
 
         to_create = []
         for shipment in shipments:
+            if shipment.warehouse_storage == shipment.warehouse_output:
+                # Do not create inventory moves
+                continue
             for move in shipment.outgoing_moves:
                 if move.state in ('cancel', 'done'):
                     continue
@@ -1117,9 +1184,8 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         'Return inventory move for the outgoing move'
         pool = Pool()
         Move = pool.get('stock.move')
-        wrh = move.shipment.warehouse
         return Move(
-            from_location=wrh.picking_location or wrh.storage_location,
+            from_location=self.warehouse_storage,
             to_location=move.from_location,
             product=move.product,
             uom=move.uom,
@@ -1134,12 +1200,14 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
 
     @classmethod
     @Workflow.transition('assigned')
+    @set_employee('assigned_by')
     def assign(cls, shipments):
         cls._sync_inventory_to_outgoing(shipments, create=True, write=False)
 
     @classmethod
     @ModelView.button
     @Workflow.transition('packed')
+    @set_employee('packed_by')
     def pack(cls, shipments):
         pool = Pool()
         Move = pool.get('stock.move')
@@ -1171,6 +1239,9 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         Move = pool.get('stock.move')
         Uom = pool.get('product.uom')
         for shipment in shipments:
+            if shipment.warehouse_storage == shipment.warehouse_output:
+                # Do not create inventory moves
+                continue
             # Sum all outgoing quantities
             outgoing_qty = {}
             for move in shipment.outgoing_moves:
@@ -1234,6 +1305,7 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('done')
+    @set_employee('done_by')
     def done(cls, shipments):
         pool = Pool()
         Move = pool.get('stock.move')
@@ -1314,6 +1386,10 @@ class ShipmentOut(Workflow, ModelSQL, ModelView):
         default = default.copy()
         default['inventory_moves'] = None
         default['outgoing_moves'] = None
+        default.setdefault('number')
+        default.setdefault('assigned_by')
+        default.setdefault('packed_by')
+        default.setdefault('done_by')
         return super(ShipmentOut, cls).copy(shipments, default=default)
 
     @classmethod
@@ -1447,6 +1523,8 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
         readonly=True)
     origins = fields.Function(fields.Char('Origins'), 'get_origins')
     number = fields.Char('Number', size=None, select=True, readonly=True)
+    received_by = employee_field("Received By")
+    done_by = employee_field("Done By")
     state = fields.Selection([
         ('draft', 'Draft'),
         ('done', 'Done'),
@@ -1588,7 +1666,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
     def get_incoming_moves(self, name):
         moves = []
         for move in self.moves:
-            if move.to_location.id == self.warehouse.input_location.id:
+            if move.to_location == self.warehouse_input:
                 moves.append(move.id)
         return moves
 
@@ -1603,7 +1681,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
     def get_inventory_moves(self, name):
         moves = []
         for move in self.moves:
-            if move.from_location.id == self.warehouse.input_location.id:
+            if move.from_location == self.warehouse_input:
                 moves.append(move.id)
         return moves
 
@@ -1673,6 +1751,9 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
         default = default.copy()
         default['inventory_moves'] = None
         default['incoming_moves'] = None
+        default.setdefault('number')
+        default.setdefault('received_by')
+        default.setdefault('done_by')
         return super(ShipmentOutReturn, cls).copy(shipments, default=default)
 
     @classmethod
@@ -1699,6 +1780,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('received')
+    @set_employee('received_by')
     def receive(cls, shipments):
         Move = Pool().get('stock.move')
         Move.do([m for s in shipments for m in s.incoming_moves])
@@ -1707,6 +1789,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('done')
+    @set_employee('done_by')
     def done(cls, shipments):
         pool = Pool()
         Move = pool.get('stock.move')
@@ -1735,7 +1818,7 @@ class ShipmentOutReturn(Workflow, ModelSQL, ModelView):
         move.uom = incoming_move.uom
         move.quantity = incoming_move.quantity
         move.from_location = incoming_move.to_location
-        move.to_location = incoming_move.shipment.warehouse.storage_location
+        move.to_location = incoming_move.shipment.warehouse_storage
         move.state = Move.default_state()
         # Product will be considered in stock only when the inventory
         # move will be made:
@@ -1817,7 +1900,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
         depends=['state'])
     planned_date = fields.Date('Planned Date',
         states={
-            'readonly': Eval('state') != 'draft',
+            'readonly': ~Eval('state').in_(['request', 'draft']),
             }, depends=['state'])
     effective_start_date = fields.Date('Effective Start Date',
         states={
@@ -1826,13 +1909,13 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
         depends=['state'])
     planned_start_date = fields.Date('Planned Start Date',
         states={
-            'readonly': ~Eval('state').in_(['draft']),
+            'readonly': ~Eval('state').in_(['request', 'draft']),
             'required': Bool(Eval('planned_date')),
             },
         depends=['state'])
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
-            'readonly': Eval('state') != 'draft',
+            'readonly': ~Eval('state').in_(['request', 'draft']),
             },
         domain=[
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
@@ -1842,18 +1925,20 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
     number = fields.Char('Number', size=None, select=True, readonly=True)
     reference = fields.Char("Reference", size=None, select=True,
         states={
-            'readonly': Eval('state') != 'draft',
+            'readonly': ~Eval('state').in_(['request', 'draft']),
             }, depends=['state'])
     from_location = fields.Many2One('stock.location', "From Location",
         required=True, states={
-            'readonly': (Eval('state') != 'draft') | Eval('moves', [0]),
+            'readonly': (~Eval('state').in_(['request', 'draft'])
+                | Eval('moves', [0])),
             },
         domain=[
             ('type', 'in', ['view', 'storage', 'lost_found']),
             ], depends=['state'])
     to_location = fields.Many2One('stock.location', "To Location",
         required=True, states={
-            'readonly': (Eval('state') != 'draft') | Eval('moves', [0]),
+            'readonly': (~Eval('state').in_(['request', 'draft'])
+                    | Eval('moves', [0])),
             }, domain=[
             ('type', 'in', ['view', 'storage', 'lost_found']),
             ], depends=['state'])
@@ -1864,10 +1949,10 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
             'readonly': (Eval('state').in_(['cancel', 'assigned', 'done'])
                 | ~Eval('from_location') | ~Eval('to_location')),
             'invisible': (Bool(Eval('transit_location'))
-                & (Eval('state') != 'draft')),
+                & ~Eval('state').in_(['request', 'draft'])),
             },
         domain=[
-            If(Eval('state') == 'draft', [
+            If(Eval('state').in_(['request', 'draft']), [
                     ('from_location', '=', Eval('from_location')),
                     ('to_location', '=', Eval('to_location')),
                     ],
@@ -1908,7 +1993,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
                 'readonly': Eval('state').in_(
                     ['assigned', 'shipped', 'done', 'cancel']),
                 'invisible': (~Eval('transit_location')
-                    | (Eval('state') == 'draft')),
+                    | Eval('state').in_(['request', 'draft'])),
                 },
             depends=['from_location', 'to_location', 'transit_location',
                 'state']),
@@ -1926,12 +2011,16 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
             states={
                 'readonly': Eval('state').in_(['done', 'cancel']),
                 'invisible': (~Eval('transit_location')
-                    | (Eval('state') == 'draft')),
+                    | Eval('state').in_(['request', 'draft'])),
                 },
             depends=['from_location', 'to_location', 'transit_location',
                 'state']),
         'get_incoming_moves', setter='set_moves')
+    assigned_by = employee_field("Received By")
+    shipped_by = employee_field("Shipped By")
+    done_by = employee_field("Done By")
     state = fields.Selection([
+            ('request', 'Request'),
             ('draft', 'Draft'),
             ('cancel', 'Canceled'),
             ('waiting', 'Waiting'),
@@ -1949,6 +2038,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
                     'before deletion.'),
                 })
         cls._transitions |= set((
+                ('request', 'draft'),
                 ('draft', 'waiting'),
                 ('waiting', 'waiting'),
                 ('waiting', 'assigned'),
@@ -1957,6 +2047,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
                 ('shipped', 'done'),
                 ('waiting', 'draft'),
                 ('assigned', 'waiting'),
+                ('request', 'cancel'),
                 ('draft', 'cancel'),
                 ('waiting', 'cancel'),
                 ('assigned', 'cancel'),
@@ -1968,10 +2059,13 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
                         ['cancel', 'shipped', 'done']),
                     },
                 'draft': {
-                    'invisible': ~Eval('state').in_(['cancel', 'waiting']),
+                    'invisible': ~Eval('state').in_(
+                        ['cancel', 'request', 'waiting']),
                     'icon': If(Eval('state') == 'cancel',
                         'tryton-clear',
-                        'tryton-go-previous'),
+                        If(Eval('state') == 'request',
+                            'tryton-go-next',
+                            'tryton-go-previous')),
                     },
                 'wait': {
                     'invisible': ~Eval('state').in_(['assigned', 'waiting',
@@ -2071,7 +2165,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
         pool = Pool()
         Config = pool.get('stock.configuration')
         if self.planned_date != self.planned_start_date:
-            return Config(1).shipment_internal_transit.id
+            return Config(1).get_multivalue('shipment_internal_transit').id
 
     @fields.depends('planned_date', 'from_location', 'to_location')
     def on_change_with_planned_start_date(self, pattern=None):
@@ -2157,6 +2251,9 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
         default['incoming_moves'] = None
         default['moves'] = None
         default.setdefault('number')
+        default.setdefault('assigned_by')
+        default.setdefault('shipped_by')
+        default.setdefault('done_by')
         copies = super(ShipmentInternal, cls).copy(shipments, default=default)
         for shipment, copy in zip(shipments, copies):
             Move.copy(shipment.outgoing_moves, default={
@@ -2176,6 +2273,8 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
         to_delete = []
         to_save = []
         for shipment in shipments:
+            if not shipment.transit_location:
+                continue
             product_qty = defaultdict(lambda: 0)
             for move in shipment.outgoing_moves:
                 if move.state == 'cancel':
@@ -2204,6 +2303,31 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
         if to_delete:
             Move.delete(to_delete)
 
+    @classmethod
+    def _set_transit(cls, shipments):
+        pool = Pool()
+        Move = pool.get('stock.move')
+
+        to_write = []
+        for shipment in shipments:
+            if not shipment.transit_location:
+                continue
+            moves = [m for m in shipment.moves
+                if m.state != 'done'
+                and m.from_location != shipment.transit_location
+                and m.to_location != shipment.transit_location]
+            Move.copy(moves, default={
+                    'from_location': shipment.transit_location.id,
+                    'planned_date': shipment.planned_date,
+                    })
+            to_write.append(moves)
+            to_write.append({
+                    'to_location': shipment.transit_location.id,
+                    'planned_date': shipment.planned_start_date,
+                    })
+        if to_write:
+            Move.write(*to_write)
+
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
@@ -2231,48 +2355,28 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
 
         Move.draft([m for s in shipments for m in s.moves])
 
-        direct = []
-        transit = []
-        for shipment in shipments:
-            if not shipment.transit_location:
-                direct.append(shipment)
-            else:
-                transit.append(shipment)
-
         moves = []
-        for shipment in direct:
+        for shipment in shipments:
+            if shipment.transit_location:
+                continue
             for move in shipment.moves:
                 if move.state != 'done':
                     move.planned_date = shipment.planned_date
                     moves.append(move)
         Move.save(moves)
 
-        to_write = []
-        for shipment in transit:
-            moves = [m for m in shipment.moves
-                if m.state != 'done'
-                and m.from_location != shipment.transit_location
-                and m.to_location != shipment.transit_location]
-            Move.copy(moves, default={
-                    'from_location': shipment.transit_location.id,
-                    'planned_date': shipment.planned_date,
-                    })
-            to_write.append(moves)
-            to_write.append({
-                    'to_location': shipment.transit_location.id,
-                    'planned_date': shipment.planned_start_date,
-                    })
-        if to_write:
-            Move.write(*to_write)
-        cls._sync_moves(transit)
+        cls._set_transit(shipments)
+        cls._sync_moves(shipments)
 
     @classmethod
     @Workflow.transition('assigned')
+    @set_employee('assigned_by')
     def assign(cls, shipments):
         pass
 
     @classmethod
     @Workflow.transition('shipped')
+    @set_employee('shipped_by')
     def ship(cls, shipments):
         pool = Pool()
         Move = pool.get('stock.move')
@@ -2286,6 +2390,7 @@ class ShipmentInternal(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('done')
+    @set_employee('done_by')
     def done(cls, shipments):
         pool = Pool()
         Move = pool.get('stock.move')
